@@ -73,8 +73,7 @@ class NetAppNVMeStorageLibrary(
     ALLOWED_SUBSYSTEM_HOST_TYPES = ['aix', 'linux', 'vmware', 'windows']
     DEFAULT_NAMESPACE_OS = 'linux'
     DEFAULT_HOST_TYPE = 'linux'
-    DEFAULT_FILTER_FUNCTION = 'capabilities.utilization < 70 and ' \
-                              'capabilities.total_volumes < 1024'
+    DEFAULT_FILTER_FUNCTION = 'capabilities.utilization < 70'
     DEFAULT_GOODNESS_FUNCTION = '100 - capabilities.utilization'
     REQUIRED_CMODE_FLAGS = ['netapp_vserver']
     NVME_PORT = 4420
@@ -151,7 +150,11 @@ class NetAppNVMeStorageLibrary(
     def _update_ssc(self):
         """Refresh the storage service catalog with the latest set of pools."""
 
-        self.ssc_library.update_ssc(self._get_flexvol_to_pool_map())
+        """Refresh the storage service catalog with the latest set of pools."""
+        if self.configuration.netapp_disaggregated_platform:
+            self.ssc_library.update_ssc_asa(self._get_cluster_to_pool_map())
+        else:
+            self.ssc_library.update_ssc(self._get_flexvol_to_pool_map())
 
     def _get_flexvol_to_pool_map(self):
         """Get the flexvols that match the pool name search pattern.
@@ -182,16 +185,25 @@ class NetAppNVMeStorageLibrary(
 
         return pools
 
+    def _get_cluster_to_pool_map(self):
+        return dot_utils.get_cluster_to_pool_map(self.client)
+
     def check_for_setup_error(self):
         """Check that the driver is working and can communicate.
 
         Discovers the namespaces on the NetApp server.
         """
-        if not self._get_flexvol_to_pool_map():
+        if (not self.configuration.netapp_disaggregated_platform
+                and not self._get_flexvol_to_pool_map()):
             msg = _('No pools are available for provisioning volumes. '
                     'Ensure that the configuration option '
                     'netapp_pool_name_search_pattern is set correctly.')
             raise na_utils.NetAppDriverException(msg)
+        elif self.configuration.netapp_disaggregated_platform:
+            if not self._get_cluster_to_pool_map():
+                msg = _('No pools are available for provisioning volumes. '
+                        'Ensure ASA r2 configuration option is set correctly.')
+                raise na_utils.NetAppDriverException(msg)
         self._add_looping_tasks()
 
         if self.namespace_ostype not in self.ALLOWED_NAMESPACE_OS_TYPES:
@@ -415,7 +427,11 @@ class NetAppNVMeStorageLibrary(
         """Extracts the namespace from API and populates the table."""
 
         path = namespace['Path']
-        (_rest, _splitter, name) = path.rpartition('/')
+        if self.configuration.netapp_disaggregated_platform:
+            name = path
+        else:
+            (_rest, _splitter, name) = path.rpartition('/')
+
         handle = self._create_namespace_handle(namespace)
         size = namespace['Size']
         return NetAppNamespace(handle, name, size, namespace)
@@ -506,7 +522,8 @@ class NetAppNVMeStorageLibrary(
 
         # Utilization and performance metrics require cluster-scoped
         # credentials
-        if self.using_cluster_credentials:
+        if (self.using_cluster_credentials
+                and not self.configuration.netapp_disaggregated_platform):
             # Get up-to-date node utilization metrics just once
             self.perf_library.update_performance_cache(ssc)
 
@@ -526,8 +543,8 @@ class NetAppNVMeStorageLibrary(
 
             # Add driver capabilities and config info
             pool['QoS_support'] = False
-            pool['multiattach'] = False
-            pool['online_extend_support'] = False
+            pool['multiattach'] = True
+            pool['online_extend_support'] = True
             pool['consistencygroup_support'] = True
             pool['consistent_group_snapshot_enabled'] = True
             pool['reserved_percentage'] = self.reserved_percentage
@@ -535,8 +552,11 @@ class NetAppNVMeStorageLibrary(
                 self.max_over_subscription_ratio)
 
             # Add up-to-date capacity info
-            capacity = self.client.get_flexvol_capacity(
-                flexvol_name=ssc_vol_name)
+            if self.configuration.netapp_disaggregated_platform:
+                capacity = self.client.get_cluster_capacity()
+            else:
+                capacity = self.client.get_flexvol_capacity(
+                    flexvol_name=ssc_vol_name)
 
             size_total_gb = capacity['size-total'] / units.Gi
             pool['total_capacity_gb'] = na_utils.round_down(size_total_gb)
@@ -559,7 +579,8 @@ class NetAppNVMeStorageLibrary(
                 pool['provisioned_capacity_gb'] = na_utils.round_down(
                     float(provisioned_cap) / units.Gi)
 
-            if self.using_cluster_credentials:
+            if (self.using_cluster_credentials and
+                    not self.configuration.netapp_disaggregated_platform):
                 dedupe_used = self.client.get_flexvol_dedupe_used_percent(
                     ssc_vol_name)
             else:
@@ -613,67 +634,46 @@ class NetAppNVMeStorageLibrary(
 
         self.namespace_table[name].size = new_size_bytes
 
-    def _get_or_create_subsystem(self, host_nqn, host_os_type):
-        """Checks for an subsystem for a host.
-
-        Creates subsystem if not already present with given host os type and
-        adds the host.
-        """
-        # Backend supports different subsystems with the same hosts, so
-        # instead of reusing non OpenStack subsystem, we make sure we only use
-        # our own, thus being compatible with custom subsystem.
-        subsystems = self.client.get_subsystem_by_host(
-            host_nqn)
-        if subsystems:
-            subsystem_name = subsystems[0]['name']
-            host_os_type = subsystems[0]['os_type']
-        else:
-            subsystem_name = na_utils.OPENSTACK_PREFIX + str(uuid.uuid4())
-            self.client.create_subsystem(subsystem_name, host_os_type,
-                                         host_nqn)
-
-        return subsystem_name, host_os_type
-
     def _find_mapped_namespace_subsystem(self, path, host_nqn):
         """Find an subsystem for a namespace mapped to the given host."""
         subsystems = [subsystem['name'] for subsystem in
                       self.client.get_subsystem_by_host(host_nqn)]
 
         # Map subsystem name to namespace-id for the requested host.
-        namespace_map = {v['subsystem']: v['uuid']
+        namespace_map = {v['uuid']: (v['subsystem_uuid'], v['subsystem'])
                          for v in self.client.get_namespace_map(path)
                          if v['subsystem'] in subsystems}
 
-        subsystem_name = n_uuid = None
+        subsystem_uuid = subsystem_name = n_uuid = None
         # Give preference to OpenStack subsystems, just use the last one if not
         # present to allow unmapping old mappings that used a custom subsystem.
-        for subsystem_name, n_uuid in namespace_map.items():
+        for n_uuid, (subsystem_uuid, subsystem_name) in namespace_map.items():
             if subsystem_name.startswith(na_utils.OPENSTACK_PREFIX):
                 break
 
-        return subsystem_name, n_uuid
+        return subsystem_uuid, subsystem_name, n_uuid
 
     def _map_namespace(self, name, host_nqn):
         """Maps namespace to the host nqn and returns its ID assigned."""
-
-        subsystem_name, subsystem_host_os = self._get_or_create_subsystem(
-            host_nqn, self.host_type)
-        if subsystem_host_os != self.host_type:
-            LOG.warning("Namespace misalignment may occur for current"
-                        " subsystem %(sub_name)s with host OS type"
-                        " %(sub_os)s. Please configure subsystem manually"
-                        " according to the type of the host OS.",
-                        {'sub_name': subsystem_name,
-                         'sub_os': subsystem_host_os})
-
         metadata = self._get_namespace_attr(name, 'metadata')
         path = metadata['Path']
         try:
-            ns_uuid = self.client.map_namespace(
-                path, subsystem_name,)
+            subsystems = self.client.get_namespace_map(path)
+            ns_uuid = subsystem_uuid = None
+            if subsystems:
+                subsystem_name = subsystems[0]['subsystem']
+                subsystem_uuid = subsystems[0]['subsystem_uuid']
+                ns_uuid = subsystems[0]['uuid']
+                self.client.map_host_with_subsystem(host_nqn, subsystem_uuid)
+            else:
+                subsystem_name = na_utils.OPENSTACK_PREFIX + str(uuid.uuid4())
+                self.client.create_subsystem(subsystem_name, self.host_type,
+                                             host_nqn)
+                ns_uuid = self.client.map_namespace(path, subsystem_name, )
             return subsystem_name, ns_uuid
         except netapp_api.NaApiError as e:
-            (subsystem_name, ns_uuid) = self._find_mapped_namespace_subsystem(
+            (_, subsystem_name, ns_uuid) =\
+                self._find_mapped_namespace_subsystem(
                 path, host_nqn)
             if ns_uuid is not None and subsystem_name:
                 return subsystem_name, ns_uuid
@@ -740,18 +740,18 @@ class NetAppNVMeStorageLibrary(
     def _unmap_namespace(self, path, host_nqn):
         """Unmaps a namespace from given host."""
 
-        namespace_unmap_list = []
-        if host_nqn:
-            (subsystem, _) = self._find_mapped_namespace_subsystem(
-                path, host_nqn)
-            namespace_unmap_list.append((path, subsystem))
-        else:
-            namespace_maps = self.client.get_namespace_map(path)
-            namespace_unmap_list = [
-                (path, m['subsystem']) for m in namespace_maps]
+        if not host_nqn:
+            LOG.warning("Nothing to unmap - host_nqn is missing: %s", path)
+            return
 
-        for _path, _subsystem in namespace_unmap_list:
-            self.client.unmap_namespace(_path, _subsystem)
+        (subsystem_uuid, _, _) = self._find_mapped_namespace_subsystem(
+            path, host_nqn)
+
+        if subsystem_uuid:
+            self.client.unmap_host_with_subsystem(host_nqn, subsystem_uuid)
+        else:
+            LOG.debug("No mapping exists between namespace: %s"
+                      " and host_nqn: %s", path, host_nqn)
 
     @coordination.synchronized('netapp-terminate-nvme-connection-{volume.id}')
     def terminate_connection(self, volume, connector, **kwargs):
